import os
import shutil
from dataclasses import dataclass
from typing import Callable, Optional, Union


@dataclass
class Dependency:
    name: str
    path: str
    required: bool


class DependenciesManager:
    """
    This class manages system-level dependencies by tracking them,
    and their status within a provided directory.
    """

    def __init__(self, dependencies_directory: str) -> None:
        """
        Initializes the Dependencies Manager.

        :param dependencies_directory: The directory containing the dependencies.
        "type dependencies_directory: str
        """
        self.dependencies_directory = dependencies_directory
        self.dependencies = []  # List to store the tracked dependencies.

    def add_dependency(self, name: str, path: str = None, required: bool = False, fallback: Optional[Callable] = None, fallback_args: tuple = None) -> Dependency:
        """
        Adds a new dependency to the tracked list.

        :param name: The name of the dependency.
        :type name: str
        :param path: Path to existing dependency. If left None, it will detect the system dependency or call fallback function, defaults to None.
        :type path: str, optional
        :param required: Indicates if the dependency is required, defaults to False.
        :type required: bool
        :param fallback: The fallback function if the dependency is not found, defaults to None.
        :type fallback: Callable, optional
        :param fallback_args: Arguments to be passed to the fallback function, defaults to ().
        :type fallback_args: tuple, optional
        :return: The newly added dependency.
        :rtype: Dependency
        """
        path = path or shutil.which(name)
        if not path or not os.path.isfile(path):
            if callable(fallback):
                if fallback_args and isinstance(fallback_args, (tuple, list)):
                    path = fallback(*fallback_args)
                elif fallback_args:
                    path = fallback(fallback_args)
        dependency = Dependency(name, path, required)
        self.dependencies.append(dependency)
        return dependency

    def get_dependency(self, name: str) -> Dependency:
        """
        Retrieves a tracked dependency by name.

        :param name: The name of the dependency.
        :type name: str
        :return: The corresponding Dependency object if it is found, None otherwise.
        :rtype: Dependency
        """
        dependency = [
            x for x in self.dependencies if x.path is not None and x.name == name]
        if dependency:
            return dependency[0]
        return None

    def has_missing_required_dependency(self) -> bool:
        """
        Checks if there is any required dependency missing.

        :return: True if a required dependency is missing, False otherwise.
        :rtype: bool
        """
        dependency = [
            x for x in self.dependencies if x.path is None and x.required]
        if dependency:
            return True
        return False

    def __getattr__(self, name: str) -> str:
        """
        Overwrites the default Python attribute getter (__getattr__).
        It allows us to get a dependency as an object attribute.

        :param name: The name of the attribute (constructed from the dependency's name).
        :type name: str
        :return: Path to dependency executable.
        :rtype: str
        """
        name = name.replace('_', '-')
        dependency = self.get_dependency(name)
        if dependency:
            return dependency.path
        return None



import os
import subprocess
from typing import Callable

from . import __APKTOOL_VERSION__, DEPENDENCIES, USER_DIRECTORIES
from .download import download_file
from .java import Java


def _download_apktool(progress_callback=None):
    path = os.path.join(USER_DIRECTORIES.user_data_dir,
                        f'apktool_{__APKTOOL_VERSION__}.jar')
    if not os.path.isfile(path):
        uri = f'https://github.com/iBotPeaches/Apktool/releases/download/v{__APKTOOL_VERSION__}/apktool_{__APKTOOL_VERSION__}.jar'
        return download_file(uri, path, progress_callback)
    return path


class Apktool:
    """
    A wrapper class for the different Apktool commands.
    """

    @staticmethod
    def advanced():
        """
        Obtain advanced information of the Apktool.

        :return: The stdout from the command execution.
        :rtype: str
        """
        return Java.run_jar(DEPENDENCIES.apktool, '--advanced')

    @staticmethod
    def version():
        """
        Obtain the version of the Apktool.

        :return: The stdout from the command execution.
        :rtype: str
        """
        return Java.run_jar(DEPENDENCIES.apktool, '--version')

    @staticmethod
    def install_framework(framework_apk, framework_path=None, tag=None):
        """
        Install framework with provided options.

        :param framework_apk: The APK framework file to be installed.
        :type framework_apk: str
        :param framework_path: The directory where to store framework files, defaults to None.
        :type framework_path: str, optional
        :param tag: The tag for the framework file, defaults to None.
        :type tag: str, optional
        :return: The stdout from the command execution.
        :rtype: str
        """
        command = f'if {framework_apk}'
        if framework_path:
            command += f' --frame-path {framework_path}'
        if tag:
            command += f' --tag {tag}'
        return Java.run_jar(DEPENDENCIES.apktool, command)

    @staticmethod
    def decode(file_apk, force=False, output=None, framework_path=None, no_res=False, no_src=False, tag=None, force_manifest=False):
        """
        Decode an APK file with provided options.

        :param file_apk: The APK file to be decoded.
        :type file_apk: str
        :param force: Whether to force delete the destination directory, defaults to False.
        :type force: bool, optional
        :param output: The name of the output folder, defaults to None.
        :type output: str, optional
        :param framework_path: The directory containing framework files, defaults to None.
        :type framework_path: str, optional
        :param no_res: Whether to not decode resources, defaults to False.
        :type no_res: bool, optional
        :param no_src: Whether to not decode sources, defaults to False.
        :type no_src: bool, optional
        :param tag: A tag for the framework file, defaults to None.
        :type tag: str, optional
        :param force_manifest: Forces decode of AndroidManifest.xml regardless of decoding of resources parameter.
        :type force_manifest: str, optional
        :return: The stdout from the command execution.
        :rtype: str
        """
        command = f'd {file_apk}'
        if force:
            command += ' --force'
        if output:
            command += f' --output {output}'
        if framework_path:
            command += f' --frame-path {framework_path}'
        if no_res:
            command += ' --no-res'
        if no_src:
            command += ' --no-src'
        if tag:
            command += f' --frame-tag {tag}'
        if force_manifest:
            command += f' --force-manifest'
        return Java.run_jar(DEPENDENCIES.apktool, command)

    @staticmethod
    def build(app_path, force_all=False, output=None, framework_path=None, use_aapt2=False):
        """
        Build the unpacked apk from the given path with the provided options.

        :param app_path: The path of the app to build.
        :type app_path: str
        :param force_all: Whether to skip changes detection and build all files, defaults to False.
        :type force_all: bool, optional
        :param output: The name of the output APK file, defaults to None.
        :type output: str, optional
        :param framework_path: The directory containing framework files, defaults to None.
        :type framework_path: str, optional
        :param use_aapt2: Whether to use aapt 2 or no, defaults to False.
        :type use_aapt2: bool, optional.
        :return: The stdout from the command execution.
        :rtype: str
        """
        command = f'b {app_path}'
        if force_all:
            command += ' --force-all'
        if output:
            command += f' --output {output}'
        if framework_path:
            command += f' --frame-path {framework_path}'
        if use_aapt2:
            command += ' --use-aapt2'
        return Java.run_jar(DEPENDENCIES.apktool, command)

    @staticmethod
    def _run(command):
        if not DEPENDENCIES.apktool:
            return
        if DEPENDENCIES.apktool.endswith('.jar'):
            return Java.run_jar(DEPENDENCIES.apktool)
        process = subprocess.run(
            f'{DEPENDENCIES.apktool} {command}', shell=True, capture_output=True, text=True)
        return process.stdout

    @staticmethod
    def install(path: str = None, progress_callback: Callable = None) -> None:
        """
        Install apktool.

        :param path: Path to existing apktool.jar. If not found, will use the system installed one or download it. Defaults to None.
        :type path: str, optional
        :param progress_callback: Callback to be called when install progress changes, defaults to None.
        :type progress_callback: callable, optional
        """
        DEPENDENCIES.add_dependency(
            'apktool', path=path, fallback=_download_apktool, fallback_args=(progress_callback,))



import importlib.util
import os

import click
import requests


@click.group()
@click.version_option()
@click.pass_context
def main(ctx):
    """
    Entry point for apkinjector application. Use as decorator @main.command() to add new terminal commands: https://click.palletsprojects.com/en/8.1.x
    """
    pass


class _PluginLoader:
    def __init__(self, plugin_folder: str) -> None:
        self.plugin_folder = plugin_folder
        self.plugins = self.load_plugins()

    def load_plugins(self):
        plugins = {}
        for filename in os.listdir(self.plugin_folder):
            if filename.endswith('.py'):
                module_name = filename[:-3]
                plugin_module = self.import_module_from_file(
                    os.path.join(self.plugin_folder, filename))
                plugins[module_name] = plugin_module
        return plugins

    def import_module_from_file(self, full_path):
        module_spec = importlib.util.spec_from_file_location(
            os.path.basename(full_path), full_path)
        module = importlib.util.module_from_spec(module_spec)
        module_spec.loader.exec_module(module)
        return module



import sys

import click

from typing import List, Tuple, Union


class ApkInjectorLogger:
    """
    Fstring like logging. e.g: log.info('Hello {}', 'World')
    """
    def __init__(self, name : str):
        """
        Initializes the logger.

        :param name: Name of the logger.
        :type name: str
        """
        self.__name__ = name

    def info(self, message: str, *args : Union[Tuple[str], List[str]]):
        """
        Prints info log message in green.

        :param message: Message to print.
        :type message: str
        """
        message = message.format(*args)
        click.echo(click.style(f'[+] {self.__name__} - {message}', fg='green'))

    def warning(self, message: str, *args : Union[Tuple[str], List[str]]):
        """
        Prints warning log message in yellow.

        :param message: Message to print.
        :type message: str
        """
        message = message.format(*args)
        click.echo(click.style(
            f'[!] {self.__name__} - {message}', fg='yellow'))

    def error(self, message: str, *args : Union[Tuple[str], List[str]]):
        """
        Prints error message in red and calls sys.exit(1).

        :param message: Message to print.
        :type message: str
        """
        message = message.format(*args)
        click.echo(click.style(f'[*] {self.__name__} - {message}', fg='red'))
        sys.exit(1)



import lzma
import os
import shutil
from typing import Callable, Union

import requests

from . import USER_DIRECTORIES
from .arch import ARCH
from .download import download_file
from .frida import Frida


class Gadget:
    """
    Frida Gadget utils.
    """
    @staticmethod
    def version() -> str:
        """
        Get recommended gadget version.

        :return: The recommended gadget version as string.
        :rtype: str
        """
        return Frida.version()

    @staticmethod
    def update(progress_callback: Callable = None) -> bool:
        """
        Downloads and updates the gadgets.

        :param progress_callback: Function to call when download progress changes, defaults to None.
        :type progress_callback: Callable, optional
        :return: True if the gadgets were successfully downloaded, False otherwise.
        :rtype: bool
        """
        version = Gadget.version()
        uri = 'https://api.github.com/repos/frida/frida/releases'
        response = requests.get(uri)
        response.raise_for_status()
        js = response.json()
        uri = None
        for release in js:
            if release['tag_name'] == version:
                uri = release['url']
        if not uri:
            return False
        response = requests.get(uri)
        response.raise_for_status()
        js = response.json()
        assets = js['assets']
        gadgets = []
        for asset in assets:
            if 'gadget' in asset['name'] and 'android' in asset['name']:
                gadgets.append(
                    [asset['name'], asset['browser_download_url']]
                )

        path = os.path.join(USER_DIRECTORIES.user_data_dir, 'gadgets', version)
        if os.path.isdir(path):
            shutil.rmtree(path)
        os.makedirs(path)
        for gadget in gadgets:
            target_path = os.path.join(path, gadget[0])
            output_path = download_file(
                gadget[1], target_path, progress_callback=progress_callback)
            with lzma.open(output_path) as _in, open(output_path.replace(".xz", ""), 'wb') as _out:
                _out.write(_in.read())
            os.remove(output_path)
        return True

    @staticmethod
    def get_gadget_path(arch: ARCH) -> Union[str, None]:
        """
        Gets the gadget path for the specified architecture.

        :param arch: The target architecture.
        :type arch: apkinjector.arch.ARCH
        :return: The path to the gadgets file if found, otherwise None.
        :rtype: Union[str, None]
        """
        version = Gadget.version()
        mappings = {
            ARCH.ARM: f'frida-gadget-{version}-android-arm.so',
            ARCH.ARM64: f'frida-gadget-{version}-android-arm64.so',
            ARCH.X64: f'frida-gadget-{version}-android-x86_64.so',
            ARCH.X86: f'frida-gadget-{version}-android-x86.so'
        }
        if arch not in mappings.keys():
            return None
        path = Gadget.get_gadget_dir()
        if not path:
            return path
        path = os.path.join(path, mappings[arch])
        if not os.path.exists(path):
            return None
        return path

    @staticmethod
    def get_gadget_dir() -> Union[str, None]:
        """
        Get the path where the gadgets are saved.

        :return: The path to the gadgets directory if found, otherwise None.
        :rtype: Union[str, None]
        """
        version = Frida.version()
        path = os.path.join(USER_DIRECTORIES.user_data_dir, 'gadgets', version)
        if os.path.isdir(path):
            return path
        return None



import subprocess
from typing import Callable, Union

from . import DEPENDENCIES, USER_DIRECTORIES, utils
from .arch import ARCH
from .download import download_file

import os
import platform
import zipfile

def _download_adb(progress_callback=None):
    def _get_bin():
        if plat in ['Linux', 'Darwin']:
            exe = os.path.join(USER_DIRECTORIES.user_data_dir, 'platform-tools', 'adb')
            os.chmod(exe, 0o755) # set executable permission
            return exe
        if plat == 'Windows':
            return os.path.join(USER_DIRECTORIES.user_data_dir, 'platform-tools', 'adb')
        return None
    plat = platform.system()
    if platform.machine() not in ['x86_64', 'amd64']:
        return None
    uri = None
    if plat == 'Linux':
        uri = 'https://dl.google.com/android/repository/platform-tools_r29.0.5-linux.zip'
    if plat == 'Windows':
        uri = 'https://dl.google.com/android/repository/platform-tools_r29.0.5-windows.zip'
    if plat == 'Darwin':
        uri = 'https://dl.google.com/android/repository/platform-tools_r29.0.5-darwin.zip'
    
    path = os.path.join(USER_DIRECTORIES.user_data_dir, 'platform-tools')
    if os.path.isdir(path):
        return _get_bin()
    downloaded = download_file(uri, f'{path}.zip', progress_callback)
    with zipfile.ZipFile(downloaded, 'r') as ref:
        ref.extractall(USER_DIRECTORIES.user_data_dir)
    return _get_bin()


class Adb:
    """
    A utility class for running adb commands.
    """
    @staticmethod
    def get_architecture() -> Union[ARCH, None]:
        """
        Get phone architecture as apkpatcher.arch.ARCH.

        This function serves as a shorthand for ``apkpatcher.utils.abi_to_arch(apkpatcher.adb.Adb.get_prop('ro.product.cpu.abi'))``.

        :return: apkpatcher.arch.ARCH if adb is installed and a device is connected. Otherwise, it returns None.
        :rtype: Union[str, None]
        """
        res = Adb.get_prop('ro.product.cpu.abi')
        if not res:
            return None
        return utils.abi_to_arch(res)

    @staticmethod
    def get_prop(prop: str) -> Union[str, None]:
        """
        Calls ``adb shell getprop {prop}``.

        :param prop: Prop to retrieve
        :type prop: str
        :return: Returns prop from device if adb is installed and a device is connected. Otherwise, it returns None.
        :rtype: Union[str, None]
        """
        return Adb.shell(f'getprop {prop}')

    @staticmethod
    def wait_for_device() -> None:
        """
        Waits for a device to be connected. If adb is not installed, it will skip without waiting.
        """
        if not DEPENDENCIES.adb:
            return None
        return Adb.run(f'wait-for-device')

    @staticmethod
    def shell(command: str) -> str:
        """
        Run an adb shell command.

        :param command: Command to be ran.
        :type command: str
        :return: Output of the command if adb is installed and a device is connected. Otherwise, it returns None.
        :rtype: str
        """
        return Adb.run(f'shell {command}')

    @staticmethod
    def run(command: str) -> Union[str, None]:
        """
        Bare method to run adb commands.

        :param command: Command to be run.
        :type command: str
        :return: Output of the command if adb is installed and a device is connected. Otherwise, it returns None.
        :rtype: Union[str, None]
        """
        if not DEPENDENCIES.adb:
            return None
        try:
            process = subprocess.run(
                f'{DEPENDENCIES.adb} {command}', shell=True, capture_output=True, text=True, timeout=5)
            return process.stdout.strip()
        except subprocess.TimeoutExpired:
            return None
        
    @staticmethod
    def install(path: str = None, progress_callback: Callable = None) -> None:
        """
        Install adb tools.

        :param path: Path to existing adb executable. If not found, will use the system installed one or download it. Defaults to None.
        :type path: str, optional
        :param progress_callback: Callback to be called when install progress changes, defaults to None.
        :type progress_callback: callable, optional
        """
        DEPENDENCIES.add_dependency(
            'adb', path=path, fallback=_download_adb, fallback_args=(progress_callback,))



import glob
import os
import shutil
import subprocess

import jdk
import time

from . import __JAVA_VERSION__, DEPENDENCIES, USER_DIRECTORIES
from .download import ProgressUnknown

from typing import Callable


def _download_java(progress_callback):
    pattern = os.path.join(USER_DIRECTORIES.user_data_dir,
                           f'jdk-{__JAVA_VERSION__}*', 'bin', 'java')
    paths = glob.glob(pattern)
    if paths:
        return paths[-1]
    if progress_callback:
        pregress = ProgressUnknown('java', None)
        progress_callback(pregress)
    tmp_directory = os.path.join(USER_DIRECTORIES.user_cache_dir)
    if os.path.isdir(tmp_directory):
        shutil.rmtree(tmp_directory)
    path = jdk.install(str(__JAVA_VERSION__), jre=True,
                       path=tmp_directory)
    if path:
        shutil.move(path, USER_DIRECTORIES.user_data_dir)
        return os.path.join(USER_DIRECTORIES.user_data_dir, 'bin', 'java')
    return None

class Java:
    """
    A utility class for running Java commands.
    """

    @staticmethod
    def run_jar(jar_file : str, command : str) -> str:
        """Execute a jar file with the provided command.

        :param jar_file: Path to the jar file to be executed.
        :type jar_file: str
        :param command: Command to be passed when executing the jar file.
        :type command: str
        :return: Stdout of the command.
        :rtype: str
        """
        process = subprocess.run(
            f'{DEPENDENCIES.java} -jar {jar_file} {command}', shell=True, capture_output=True, text=True)
        return process.stdout
    
    @staticmethod
    def install(path: str = None, progress_callback: Callable = None) -> None:
        """
        Installs java

        :param path: Path to existing java if not found will use the system installed one or download it. Defaults to None.
        :type path: str, optional
        :param progress_callback: Callback to be called when install progress changes. Defaults to None.
        :type progress_callback: Callable, optional.

        """
        DEPENDENCIES.add_dependency(
            'java', path=path, fallback=_download_java, fallback_args=(progress_callback,))



from setuptools import setup, find_packages
from pathlib import Path

this_directory = Path(__file__).parent
long_description = (this_directory / "README.md").read_text()

setup(
    name='apkinjector',
    version='1.1.5',
    description='Utilities to help injecting libraries and frida in apks.',
    long_description=long_description,
    long_description_content_type='text/markdown',
    author='Marcel Alexandru Nitan',
    author_email='nitan.marcel@gmail.com',
    url='https://nitanmarcel.github.io/apkinjector',
    keywords=['FRIDA', 'APK', 'INJECTION', 'INJECT'],
    packages=find_packages(),
    entry_points={
        'console_scripts': [
            'apkinjector = apkinjector.__main__:start'
        ]
    },
    install_requires=[
        'requests',
        'appdirs',
        'install-jdk',
        'click',
        'pyelftools',
        'colorama'
    ],
    classifiers=[
        'License :: OSI Approved :: GNU General Public License v2 (GPLv2)',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',

    ]
)



import os
import shutil
from typing import List

from elftools.elf.elffile import ELFFile

from .arch import ARCH
from .utils import arch_to_abi


class Injector:
    @staticmethod
    def inject_library(source: str, unpack_path: str, smali_path: str = None, extra_files: List[str] = None, skip_copy=False) -> None:
        """
        Injects a library into an unpacked apk.

        :param source: The path to the shared library to inject.
        :type source: str
        :param unpack_path: The path to the apktool unpacked apk.
        :type unpack_path: str
        :param smali_path: The path to the smali file to inject into. Leave empty to skip this step. Defaults to None.
        :type smali_path: str, optional
        :param extra_files: Extra files to copy over the to the apk lib/ folder. Defaults to None.
        :type extra_files: list of str, optional
        :param skip_copy: Whatever to copy the library into target, or skip and just edit the smali. Defaults to False.
        :type skip_copy: bool, optional

        :return: If the injection was successful or not.
        :rtype: bool
        """
        lib, _ = os.path.splitext(os.path.basename(source))
        lib = lib.split('lib', 1)[-1]
        lines = []
        injected = False
        if smali_path and not os.path.isfile(smali_path):
            return injected

        arch = Injector.guess_arch_from_lib(source)
        abi = arch_to_abi(arch.lower())

        if not skip_copy:
            path = os.path.join(unpack_path, 'lib', abi)
            if not os.path.isdir(path):
                os.makedirs(path)
            paths = []
            paths.append(source)
            if extra_files:
                paths.extend(extra_files)
            for file in paths:
                name = os.path.basename(file)
                dest = os.path.join(path, name)
                if os.path.isfile(dest):
                    os.remove(dest)
                shutil.copyfile(file, dest)

        injected = not bool(smali_path)
        if smali_path:
            with open(smali_path, 'r') as smali:
                lines = smali.readlines()
            matches = []
            for line in lines:
                if 'const-string v0, "{}"\n'.format(lib) in line:
                    matches.append(line)
                if matches and 'invoke-static {v0}, Ljava/lang/System;->loadLibrary(Ljava/lang/String;)V\n' in line:
                    matches.append(line)
            if len(matches) >= 2:
                return True
            for line in lines:
                if line.startswith('.method public constructor <init>()V') or line.startswith('.method static constructor <clinit>()V'):
                    index = lines.index(line)
                    if lines[index + 1].startswith('    .locals'):
                        lines.insert(
                            index + 2, '    const-string v0, "{}"\n'.format(lib))
                        lines.insert(
                            index + 3, '    invoke-static {v0}, Ljava/lang/System;->loadLibrary(Ljava/lang/String;)V\n')
                        injected = True
                        break
                    else:
                        lines.insert(
                            index + 1, '    const-string v0, "{}"\n'.format(lib))
                        lines.insert(
                            index + 2, '    invoke-static {v0}, Ljava/lang/System;->loadLibrary(Ljava/lang/String;)V\n')
                        injected = True
                        break
            with open(smali_path, 'w') as smali:
                smali.writelines(lines)
        return injected
    
    @staticmethod
    def guess_arch_from_lib(lib_path: str) -> ARCH:
        with open(lib_path, 'rb') as libfile:
            elffile = ELFFile(libfile)
            arch = elffile.get_machine_arch()
            if arch.lower() in ['aarch64', 'arm64']:
                arch = ARCH.ARM64
            if arch.lower() in ['x64']:
                arch = ARCH.X64
        return arch
            